@@ -1,9 +1,9 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997,2005 Sergey Poznyakoff
+   Copyright (C) 1997, 2005, 2006, 2007 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
  
    GNU cflow is distributed in the hope that it will be useful,
@@ -30,19 +30,20 @@ typedef struct {
 void parse_declaration(Ident*, int);
 void parse_variable_declaration(Ident*, int);
 void parse_function_declaration(Ident*, int);
-void parse_dcl(Ident*);
+void parse_dcl(Ident*, int maybe_knr);
 void parse_knr_dcl(Ident*);
 void parse_typedef();
 void expression();
 void initializer_list();
 void func_body();
-void declare(Ident*);
+void declare(Ident*, int maybe_knr);
 void declare_type(Ident*);
 int dcl(Ident*);
 int parmdcl(Ident*);
 int dirdcl(Ident*);
 void skip_struct();
 Symbol *get_symbol(char *name);
+Symbol *install_ident(char *name);
 void maybe_parm_list(int *parm_cnt_return);
     
 void call(char*, int);
@@ -539,7 +540,7 @@ parse_variable_declaration(Ident *ident, int parm)
 	  }
      }
  again:
-     parse_dcl(ident);
+     parse_dcl(ident, 0);
      
  select:    
      switch (tok.type) {
@@ -612,58 +613,7 @@ void
 parse_knr_dcl(Ident *ident)
 {
      ident->type_end = -1;
-     parse_dcl(ident);
-     if (strict_ansi)
-	  return;
-     
-     switch (tok.type) {
-     case IDENTIFIER:
-     case TYPE:
-     case STRUCT:
-	  if (ident->parmcnt >= 0) {
-	       /* maybe K&R function definition */
-	       int parmcnt, stop;
-	       Stackpos sp, new_sp;
-	       Ident id;
-	       
-	       mark(sp);
-	       parmcnt = 0;
-	       
-	       for (stop = 0; !stop && parmcnt < ident->parmcnt;
-		    nexttoken()) {
-		    id.type_end = -1;
-		    switch (tok.type) {
-		    case LBRACE:
-		    case LBRACE0:
-			 putback();
-			 stop = 1;
-			 break;
-		    case TYPE:
-		    case IDENTIFIER:
-		    case STRUCT:
-			 putback();
-			 mark(new_sp);
-			 if (dcl(&id) == 0) {
-			      parmcnt++;
-			      if (tok.type == ',') {
-				   do {
-					tos = id.type_end; /* ouch! */
-					restore(new_sp);
-					dcl(&id);
-				   } while (tok.type == ',');
-			      } else if (tok.type != ';')
-				   putback();
-			      break;
-			 }
-			 /* else */
-			 /* FALLTHRU */
-		    default:
-			 restore(sp);
-			 return;
-		    }
-	       }
-	  }
-     }
+     parse_dcl(ident, !strict_ansi);     
 }
 
 void
@@ -716,7 +666,7 @@ parse_typedef()
 }
 
 void
-parse_dcl(Ident *ident)
+parse_dcl(Ident *ident, int maybe_knr)
 {
      ident->parmcnt = -1;
      ident->name = NULL;
@@ -724,7 +674,7 @@ parse_dcl(Ident *ident)
      dcl(ident);
      save_stack();
      if (ident->name)
-	  declare(ident);
+	  declare(ident, maybe_knr);
      else 
 	  undo_save_stack();
 }
@@ -748,8 +698,7 @@ dcl(Ident *idptr)
 	       else if (type != MODIFIER) 
 		    break;
 	  } else if (tok.type == ')') {
-	       putback();
-	       break;
+	       return 1;
 	  }
      }
      if (idptr && idptr->type_end == -1)
@@ -947,8 +896,61 @@ func_body()
      }
 }
 
+int
+get_knr_args(Ident *ident)
+{
+     int parmcnt, stop;
+     Stackpos sp, new_sp;
+     Ident id;
+
+     switch (tok.type) {
+     case IDENTIFIER:
+     case TYPE:
+     case STRUCT:
+	  /* maybe K&R function definition */
+	  
+	  mark(sp);
+	  parmcnt = 0;
+	  
+	  for (stop = 0; !stop && parmcnt < ident->parmcnt;
+	       nexttoken()) {
+	       id.type_end = -1;
+	       switch (tok.type) {
+	       case LBRACE:
+	       case LBRACE0:
+		    putback();
+		    stop = 1;
+		    break;
+	       case TYPE:
+	       case IDENTIFIER:
+	       case STRUCT:
+		    putback();
+		    mark(new_sp);
+		    if (dcl(&id) == 0) {
+			 parmcnt++;
+			 if (tok.type == ',') {
+			      do {
+				   tos = id.type_end; /* ouch! */
+				   restore(new_sp);
+				   dcl(&id);
+			      } while (tok.type == ',');
+			 } else if (tok.type != ';')
+			      putback();
+			 break;
+		    }
+		    /* else */
+		    /* FALLTHRU */
+	       default:
+		    restore(sp);
+		    return 1;
+	       }
+	  }
+     }
+     return 0;
+}
+
 void
-declare(Ident *ident)
+declare(Ident *ident, int maybe_knr)
 {
      Symbol *sp;
      
@@ -965,8 +967,10 @@ declare(Ident *ident)
 	  sp->arity = -1;
 	  return;
      } 
-     
-     if ((ident->parmcnt >= 0 && !(tok.type == LBRACE || tok.type == LBRACE0))
+
+     if ((ident->parmcnt >= 0
+	  && (!maybe_knr || get_knr_args(ident) == 0)
+	  && !(tok.type == LBRACE || tok.type == LBRACE0 || tok.type == TYPE))
 	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)) {
 	  undo_save_stack();
 	  /* add_external()?? */
@@ -975,11 +979,16 @@ declare(Ident *ident)
      
      sp = get_symbol(ident->name);
      if (sp->source) {
-	  error_at_line(0, 0, filename, ident->line, 
-			_("%s/%d redefined"),
-			ident->name, sp->arity);
-	  error_at_line(0, 0, sp->source, sp->def_line,
-			_("this is the place of previous definition"));
+	  if (ident->storage == StaticStorage
+	      && (sp->storage != StaticStorage || level > 0)) {
+	       sp = install_ident(ident->name);
+	  } else {
+	       error_at_line(0, 0, filename, ident->line, 
+			     _("%s/%d redefined"),
+			     ident->name, sp->arity);
+	       error_at_line(0, 0, sp->source, sp->def_line,
+			     _("this is the place of previous definition"));
+	  }
      }
 
      sp->type = SymIdentifier;
@@ -1023,18 +1032,10 @@ declare_type(Ident *ident)
 }
 
 Symbol *
-get_symbol(char *name)
+install_ident(char *name)
 {
      Symbol *sp;
-     
-     if (sp = lookup(name)) {
-	  for (; sp; sp = sp->next) {
-	       if (sp->type == SymIdentifier && strcmp(sp->name, name) == 0)
-		    break;
-	  }
-	  if (sp)
-	       return sp;
-     }
+
      sp = install(name);
      sp->type = SymIdentifier;
      sp->arity = -1;
@@ -1048,13 +1049,30 @@ get_symbol(char *name)
      return sp;
 }
 
+Symbol *
+get_symbol(char *name)
+{
+     Symbol *sp;
+     
+     if (sp = lookup(name)) {
+	  for (; sp; sp = sp->next) {
+	       if (sp->type == SymIdentifier && strcmp(sp->name, name) == 0)
+		    break;
+	  }
+	  if (sp)
+	       return sp;
+     }
+     return install_ident(name);
+}
+
 Symbol *
 add_reference(char *name, int line)
 {
      Symbol *sp = get_symbol(name);
      Ref *refptr;
 
-     if (sp->storage == AutoStorage)
+     if (sp->storage == AutoStorage
+	 || (sp->storage == StaticStorage && globals_only()))
 	  return NULL;
      refptr = xmalloc(sizeof(*refptr));
      refptr->source = filename;
